#include <stdio.h>
#include <fcntl.h>

#ifdef __linux__
#include <unistd.h>
#include <sys/mman.h>
#else
#include <windows.h>
#endif

#include <qapplication.h>
#include <qlistbox.h>
#include <qlineedit.h>
#include <qfiledialog.h>
#include <qmessagebox.h>
#include <qprogressdialog.h>
#include <qlabel.h>
#include <qdom.h>
#include <qtextbrowser.h>
#include <qclipboard.h>
#include <qsocketnotifier.h>
#include <qpushbutton.h>
#include <qcombobox.h>
#include <qmenubar.h>
#include <qlayout.h>
#include <qthread.h>
#include <qfileinfo.h>
#include <qtextcodec.h>
#include <list>
#include <vector>
#include "textbrowser.h"
#include "indexer.h"

// due to different building models
#ifdef __linux__
#include "moc_pingvo.h"
#else
#include "pingvo.h"
#endif

#include "SoundArchive.h"

#ifdef __linux__
extern int recorder_datafd;
void recorder_init();
void recorder_cleanup();
void do_recording();
void recorder_process_replies();
#endif

IdxView *idxview = 0;

ISoundArchive *sndArc;
void *snddll;
QString sndarcpath;

IdxViewItem::IdxViewItem( char *ittext0 ): ittext(ittext0), refcount(0) {
}

QString IdxViewItem::text() const {
  return QString::fromUtf8( ittext );
}

IdxView::IdxView( QWidget *parent, const char *name, WFlags f):
	IdxViewUI( parent, name, f ),
	status_bar(0),
	dloader( this ),
	configChanged(false), ign_dict_toggle(false),
	dont_do_lookup(false),
	winX(-1), winY(-1), winW(-1), winH(-1),
	confX(-1), confY(-1), confWidth(-1), confHeight(-1),
	defArtX(0), defArtY(0), defArtWidth(420), defArtHeight(470),
	configSaveTimer(0), configured(false),
	dict1(-1), dict2(-1), dict3(-1) {

  loadConfig();
  show();
  qApp->processEvents();
  if ( winX != -1 && winY != -1 && winW != -1 && winH != -1 ) {
	move( winX, winY );
	resize( winW, winH );
  }
#ifdef __linux__
  recorder_init();
  ev_sn = new QSocketNotifier( recorder_datafd, QSocketNotifier::Read, this );
  connect( ev_sn, SIGNAL(activated(int)), SLOT(xRecordEventsAvail()) );
  do_recording();
#endif
  configured = true;

  QMenuBar *mb = new QMenuBar( this );
  if ( layout() ) layout()->setMenuBar( mb );
  mb->show();
  dictMenu = new QPopupMenu( mb );
  mb->insertItem( "Dictionaries", dictMenu );
  rebuildDictMenu();
  connect( dictMenu, SIGNAL(activated(int)), SLOT(dictMenuAction(int)) );
}

IdxView::~IdxView() {
  dloader.stopProcessing = true;
  while ( dloader.running() ) {
	qApp->processEvents();
	usleep( 300000 );	
  }
#ifdef __linux__
  delete ev_sn;
  recorder_cleanup();
#endif
}

lookupEntry *IdxView::lookupNode( QString text, bool partial_ok, bool longest_match ) {
  int dict;
  return lookupNode( text, dict, partial_ok, longest_match );
}

lookupEntry *IdxView::lookupNode( QString text, int& out_dict,
	bool partial_ok, bool longest_match ) {
  lookupEntry *le, *max_le = 0;
  int match_len, max_match = 0;

  out_dict = -1;
  for ( unsigned int i = 0; i < dictionaries.size(); i++ ) {
  	le = dictionaries[i]->lookupNode( text, partial_ok, &match_len );
  	if ( le ) {
		if ( !partial_ok || !longest_match ) {
			out_dict = i;
			return le;
		}
		if ( max_match < match_len ) {
			out_dict = i;
			max_le = le;
			max_match = match_len;
		}
	}
  }

  return max_le;
}

QString win2uni( char *win ) {
  QTextCodec *codec = QTextCodec::codecForName( "CP1251" );

  if ( !codec ) return win;
  return codec->toUnicode( win );
}

QString ArticleXML::getRichText() {
  QString text;

  text = "<table border=0 cellspacing=0 cellpadding=0>"
	  "<tr><td><h2>" + name + "&nbsp;";
  if ( sndArc ) {
	QString sndWord = name.lower() + ".wav";
	if ( sndArc->FindWord( sndWord.ucs2() ) )
	  text += "<a name=\"#" + sndWord + "\" href=\"#\"><img src=\"audio.bmp\" width=16 height=16></a>";
  }

  first_m0 = true;
  text += "</h2></td></tr>" + xmlToRichText( e ) + "</table>";

  return text;
}

QString ArticleXML::xmlToRichText( QDomElement e ) {
  QDomNode n;
  QString text, ref;
  QChar c;
  
  for ( n = e.firstChild(); !n.isNull(); n = n.nextSibling() ) {
  	if ( n.isText() )
		text += n.toText().data();
	else if ( n.isElement() ) {
		e = n.toElement();
		if ( e.tagName() == "ref" ) {
			QString ref2 = xmlToRichText( e );
			if ( ref.isNull() ) ref = ref2;
			else ref += " " + ref2;
		
			c = ref.at(ref.length()-1);
  			if ( c == ',' || c == ';' ) {
				text += buildRichRef( ref );
				ref = QString::null;
			}
			continue;
		}
		if ( !ref.isNull() ) {
			text += buildRichRef( ref );
			ref = QString::null;
		}
		if ( e.tagName() == "m0" ) {
			if ( !first_m0 )
				text += "<tr><td>" + xmlToRichText( e ) +
					"</td></tr>";
			else
				first_m0 = false;
		}
		else if ( e.tagName() == "m1" )
			text += "<tr><td>&nbsp;" + xmlToRichText( e ) +
				"</td></tr>";
		else if ( e.tagName() == "m2" )
			text += "<tr><td>&nbsp;&nbsp;" + xmlToRichText( e ) +
				"</td></tr>";
		else if ( e.tagName() == "m3" )
			text += "<tr><td>&nbsp;&nbsp;&nbsp;" + xmlToRichText( e ) +
				"</td></tr>";
		else if ( e.tagName() == "com" )
			text += "<i><font color=\"#007F00\">" +
				xmlToRichText( e ) + "</font></i>\n";
		else if ( e.tagName() == "ex" )
			text += "<font color=\"#7F7F7F\">" +
				xmlToRichText( e ) + "</font>\n";
		else if ( e.tagName() == "trn" )
			text += xmlToRichText( e );
		else if ( e.tagName() == "b" )
			text += "<b>" + xmlToRichText( e ) + "</b>";
		else if ( e.tagName() == "t" ) {
			char *s, *s1, *p, buf[30];
			int c;
			QString t = xmlToRichText(e);

			s = (char*) t.latin1();
			if ( s ) {
				p = buf;
				for (;;) {
					c = strtol( s, &s1, 10 );
					if ( s == s1 ) break;
					
					*p++ = c;

					s = s1;
					if ( *s == ' ' ) s++;
				}
				*p = 0;

				text += "<font face=\"Lingvo OEM\" size=5>[";
				text += win2uni( buf );
				text += "]</font> ";
			}
		}
		else
			printf( "element with tag: %s\n", e.tagName().latin1() );
	}
  }

  if ( !ref.isNull() ) text += buildRichRef( ref );
  return text;
}

QString ArticleXML::buildRichRef( QString ref ) {
  QString atxt;
  QChar c;

  c = ref.at(ref.length()-1);
  if ( c == ',' || c == ';' ) {
	ref = ref.left( ref.length()-1 );
	atxt = c;
  }
  return "<a name=\"" + ref +"\" href=\"#\">" + ref + "</a>" + atxt + " ";
}

QString ArticleDsl::getRichText() {
  QString text, art, tag, ref;
  int state;

  text = "<table border=0 cellspacing=0 cellpadding=0>"
	  "<tr><td><h2>" + name + "</h2></td></tr>";
  
  state = 0;
  for ( int i = 0; i < s.length(); i++ ) {
  	switch ( state ) {
	case 0:
		if ( s[i] == '[' ) {
			tag = "";
			state = 1;
		}
		else if ( s[i] == '\\' )
			state = 2;
		else 
			art += s[i];
		break;
	case 1:
		if ( s[i] == ']' ) {
			if ( tag == "/m" ) art += "</td></tr>";
			else if ( tag == "m1" ) art += "<tr><td>&nbsp;";
			else if ( tag == "m2" ) art += "<tr><td>&nbsp;&nbsp;";
			else if ( tag == "com" )
				art += "<i><font color=\"#007F00\">";
			else if ( tag == "/com" ) art += "</font></i>\n";
			else if ( tag == "ex" )
				art += "<font color=\"#7F7F7F\">";
			else if ( tag == "/ex" ) art += "</font>\n";
			else if ( tag == "ref" ) {
				ref = "";
				state = 3;
				break;
			}
			else if ( tag == "/ref" )
				art += "<a name=\"" + ref +"\" href=\"#\">" + ref + "</a>";
			state = 0;
		}
		else tag += s[i];
		break;
	case 2:
		art += s[i];
		state = 0;
		break;
	case 3:
		if ( s[i] == '[' ) {
			tag = "";
			state = 1;
		}
		else ref += s[i];
		break;
	}
  }
  
  text += art + "</table>";

#if 0
  puts( "source" );
  puts( s.latin1() );
  puts( "" );
  puts( "rich text" );
  puts( text.latin1() );
  puts( "" );
#endif

  return text;
}

Dictionary::Dictionary(): rootNode(0) {
}

lookupEntry *Dictionary::lookupNode( QString text, bool partial_ok, int *match_len ) {
  unsigned int i;
  int j, sz;
  short c;
  lookupEntry *ent, *c_ent;

  if ( !rootNode ) return 0;

  ent = 0;
  for ( i = 0; i < text.length(); i++ ) {
  	if ( cachedText.at(i) != text.at(i) ) break;
	ent = cachedNodes[i];
  }
  if ( !ent ) ent = rootNode;
  
  cachedNodes.resize( text.length() );
  cachedText.truncate( text.length() );

  for (; i < text.length(); i++ ) {
  	c = text.at(i).unicode();
	sz = ent->dchilds_count;
	for ( j = 0; j < sz; j++ ) {
		c_ent = (lookupEntry*)(tree_start + ent->child_offs[j]);
	  	if ( c_ent->c == c ) {
			ent = c_ent;
			cachedText.at(i) = c;
			cachedNodes[i] = ent;
			break;
		}
	}
	if ( j == sz ) {
		cachedText.truncate( i );
		if ( !partial_ok ) return 0;
		break;
	}
  }
  if ( match_len ) *match_len = i;

  return ent;
}

char *Dictionary::itemText( int num ) {
  return (char*)words_start + ((int*)words_start)[num];
}

void Dictionary::getArticle( QByteArray& buf, const QString& word ) {
  int idxrec[2];

  buf.resize( 0 );

  lookupEntry *ent = lookupNode( word, false );
  if ( !ent || ent->word_number == -1 ) return;

  QFile f( idxname ), f2( dname );

  if ( !f.open( IO_ReadOnly ) ) {
  	puts( "pingvo: failed to open dictionary index" );
	return;
  }

  f.at( ent->word_number * sizeof idxrec );
  f.readBlock( (char*)idxrec, sizeof idxrec );
  f.close();
  
  if ( !f2.open( IO_ReadOnly ) ) {
  	puts( "pingvo.cc failed to open dictionary" );
	return;
  }
  f2.at( idxrec[0] );
  buf.resize( idxrec[1] );
  f2.readBlock( buf.data(), idxrec[1] );
  f2.close();

  return;
}

Article *DictionaryXML::getArticle( QString word ) {
  QDomDocument doc("dictionary");
  QByteArray buf;
  Dictionary *dict;
  int len;

  Dictionary::getArticle( buf, word );

  len = buf.size();
  if ( !len ) return 0;

  buf.resize( len + 1 );
  buf[len] = 0;

  doc.setContent( buf );
  return new ArticleXML( word, doc.documentElement() );
}

Article *DictionaryDsl::getArticle( QString word ) {
  QByteArray buf;
  int len;
  Dictionary *dict;

  Dictionary::getArticle( buf, word );
  
  len = buf.size();
  if ( !len ) return 0;

  buf.resize( len + 2 );
  buf[len] = 0;
  buf[len+1] = 0;

  return new ArticleDsl( word, QString::fromUcs2( (unsigned short*)buf.data() ) );
}

void IdxView::initDictXML( Dictionary *ndict, QString fname ) {
  int pos;

  ndict->state = Dictionary::offline;
  ndict->dname = fname;
  ndict->idxname = fname.left( fname.length() - 4 ) + ".idx";
  ndict->wfname = fname.left( fname.length() - 4 ) + ".words";
  ndict->treename = fname.left( fname.length() - 4 ) + ".tree";

  pos = fname.findRev( '/' );
  if ( pos ) ndict->name = fname.right( fname.length() - pos - 1 );
  else ndict->name = fname;
}

LoaderTask::LoaderTask( taskType task0, int dict_idx0 ):
	task(task0), dict_idx(dict_idx0) {
}

DictionaryLoader::DictionaryLoader( IdxView *parent ): idxview(parent),
	stopProcessing(false) {
}

void DictionaryLoader::loadDictXML( int dict_idx ) {
  idxview->dictionaries[dict_idx]->state = Dictionary::loading;
  tlLock.lock();
  taskList.push_back( new LoaderTask( LoaderTask::loadDict, dict_idx ) );
  tlLock.unlock();
  if ( !running() ) start();
}

void DictionaryLoader::unloadDictXML( int dict_idx ) {
  idxview->dictionaries[dict_idx]->state = Dictionary::unloading;
  tlLock.lock();
  taskList.push_back( new LoaderTask( LoaderTask::unloadDict, dict_idx ) );
  tlLock.unlock();
  if ( !running() ) start();
}

void DictionaryLoader::run() {
  LoaderTask *ldtask;
  while ( !stopProcessing ) {
  	tlLock.lock();
	if ( !taskList.size() ) {
		tlLock.unlock();
		break;
	}
	ldtask = taskList.front();
	taskList.pop_front();
	tlLock.unlock();
	if ( ldtask->task == LoaderTask::loadDict )
		loadDictXML_impl( ldtask->dict_idx );
	else if ( ldtask->task == LoaderTask::unloadDict )
		unloadDictXML_impl( ldtask->dict_idx );
	delete ldtask;
  }
}

void DictionaryLoader::reportEvent( char ev ) {
  char *buf = new char;
  *buf = ev;

  ackLock.lock();
  qApp->postEvent( idxview, new QCustomEvent( (QEvent::Type)(QEvent::User + 2234), buf ) );
  ack.wait( &ackLock );
  ackLock.unlock();
}

bool less_loadedItem ( DictionaryLoader::loadedItem *it1,
	DictionaryLoader::loadedItem *it2 ) {
  return it1->it->text() < it2->it->text();
}

#ifdef __linux__
void DictionaryLoader::map_words( Dictionary *dict ) {
  int fdwords = open( dict->wfname.latin1(), O_RDONLY );
  dict->wordslen = lseek( fdwords, 0, SEEK_END );
  dict->words_start = (unsigned char*) mmap( 0, dict->wordslen, PROT_READ,
  	MAP_SHARED, fdwords, 0 );
  ::close( fdwords );
}

void DictionaryLoader::map_tree( Dictionary *dict ) {
  int fdtree = open( dict->treename, O_RDONLY );
  dict->treelen = lseek( fdtree, 0, SEEK_END );
  dict->tree_start = (unsigned char*) mmap( 0, dict->treelen, PROT_READ,
  	MAP_SHARED, fdtree, 0 );
  ::close( fdtree );
}

#else

void DictionaryLoader::map_words( Dictionary *dict ) {
  HANDLE hFile, hMap;
  hFile = CreateFile( dict->wfname.ucs2(), GENERIC_READ, FILE_SHARE_READ,
		NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
  dict->wordslen = GetFileSize( hFile, NULL );
  hMap = CreateFileMapping( hFile, 0, PAGE_READONLY, 0, 0, NULL );
  dict->words_start = (unsigned char*) MapViewOfFile( hMap, FILE_MAP_READ, 0, 0, 0 ); 
  CloseHandle( hFile );
  CloseHandle( hMap );
}

void DictionaryLoader::map_tree( Dictionary *dict ) {
  HANDLE hFile, hMap;
  hFile = CreateFile( dict->treename.ucs2(), GENERIC_READ, FILE_SHARE_READ,
		NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
  dict->treelen = GetFileSize( hFile, NULL );
  hMap = CreateFileMapping( hFile, 0, PAGE_READONLY, 0, 0, NULL );
  dict->tree_start = (unsigned char*) MapViewOfFile( hMap, FILE_MAP_READ, 0, 0, 0 ); 
  CloseHandle( hFile );
  CloseHandle( hMap );
}
#endif

void DictionaryLoader::loadDictXML_impl( int dict_idx ) {
  IdxViewItem *node;
  loadedItem *litem;
  Dictionary *dict;
  char *word;
  QString it_name;
  FILE *fidx;
  int wcount;

  dict = idxview->dictionaries[dict_idx];

  fidx = fopen( dict->idxname.latin1(), "r" );
  fseek( fidx, 0, SEEK_END );
  wcount = ftell( fidx ) / 8;
  fclose( fidx );

  map_words( dict );

//  puts( "looking for new words" );

  dict->itMap.clear();
  for ( int wnum = 0; wnum < wcount; wnum++ ) {
  	word = (char*)dict->words_start + ((int*)dict->words_start)[wnum];
	node = new IdxViewItem( word );
	litem = new loadedItem;
	litem->it = node;
	loaded.push_back( litem );
  }

//  puts( "looking for hook up points" );

  QListBoxItem *it = idxview->lb->firstItem();
  IdxViewItem *prev = 0;
  list <DictionaryLoader::loadedItem*>::iterator ldit = loaded.begin();
  for (; ldit != loaded.end(); ++ldit ) {
	IdxViewItem *node = (*ldit)->it;

	while ( it ) {
		if ( it->text() < node->text() ) {
			prev = (IdxViewItem*) it;
			it = it->next();
			continue;
		}
		
		if ( it->text() == node->text() ) {
			delete node;
			(*ldit)->it = 0;
			node = (IdxViewItem*) it;
		}
		break;
	}
	(*ldit)->after = prev;
	prev = node;

	node->refcount++;
	dict->itMap.push_back( node );
  }

//  puts( "finished" );

  map_tree( dict );
  dict->rootNode = (lookupEntry*)dict->tree_start;

  dict->state = Dictionary::ready;
//  puts( "reporting event" );
  reportEvent( 1 );
//  puts( "dictionary is on" );
}

void DictionaryLoader::unloadDictXML_impl( int dictnum ) {
  IdxViewItem *it;
  Dictionary *dict = idxview->dictionaries[dictnum], *dict2;
  unsigned int i, j;
  lookupEntry *entry;

  dict->rootNode = 0;
#ifdef __linux__
  munmap( dict->tree_start, dict->treelen );
#else
  UnmapViewOfFile( dict->tree_start );
#endif

  for ( j = 0; j < dict->itMap.size(); j++ ) {
  	QString word;

	it = dict->itMap[j];
	word = it->text();
	it->refcount--;
	if ( it->refcount ) {
		if ( it->ittext > (char*) dict->words_start &&
			it->ittext < (char*)dict->words_start + dict->wordslen ) {
			for ( i = 0; i < idxview->dictionaries.size(); i++ ) {
				dict2 = idxview->dictionaries[i];
				entry = dict2->lookupNode( word, false );
				if ( entry && entry->word_number != -1 ) {
					it->ittext = dict2->itemText( entry->word_number );
					break;
				}
			}
		}
	}
	else
		removed.push_back( it );
  }
  dict->cachedText = QString::null;

  dict->itMap.clear();
  dict->state = Dictionary::offline;
  reportEvent( 2 );

#ifdef __linux__
  munmap( dict->words_start, dict->wordslen );
#else
  UnmapViewOfFile( dict->words_start );
#endif
}

class IndexBuilderThread: public QThread {
public:
  QString xmlname;

  IndexBuilderThread( const char *xmlname );
  void run();
};

IndexBuilderThread::IndexBuilderThread( const char *xmlname0 ) {
  xmlname = QString::fromLatin1( xmlname0 );
}

void IndexBuilderThread::run() {
  if ( xmlname.right(4).lower() == ".xml" ) {
	IndexBuilder b;
	b.build( xmlname.latin1() );
  }
  else if ( xmlname.right(4).lower() == ".dat" ) {
  	IndexBuilderDsl b;
	b.build( xmlname.latin1() );
  }
}

void IdxView::loadDictXML( int dict_idx ) {
  QStringList lst, fail;
  unsigned int i;

  lst += dictionaries[dict_idx]->dname;
  lst += dictionaries[dict_idx]->idxname;
  lst += dictionaries[dict_idx]->wfname;
  lst += dictionaries[dict_idx]->treename;
  
  for ( i = 0; i < lst.size(); i++ )
  	if ( !QFile::exists( lst[i] ) ) fail += lst[i];

  if ( fail.size() ) {
	int rc;

  	QString msgtext = "Following files are missing:\n";
	for ( i = 0; i < fail.size(); i++ )
		msgtext += fail[i] + "\n";
	msgtext += "Dictionary Indexes need to be rebuilded, Ok?";
  	rc = QMessageBox::warning( this, "Pingvo Warning", msgtext,
		"Ok", "Cancel" );
	if ( rc == 0 ) {
		QProgressDialog d( "Building Indexes...", "Cancel", 2,
			this, 0, true  );
		d.setCancelButton( 0 );
		d.setProgress( 1 );
		IndexBuilderThread b( dictionaries[dict_idx]->dname.latin1() );
		b.start();
		while ( b.running() ) {
			qApp->processEvents();
			usleep( 100000 );
		}
		d.setProgress( 2 );
	}
	else {
		rebuildDictMenu();
		updateButtons();
		return;
	}
  }

  fail.clear();
  for ( i = 0; i < lst.size(); i++ ) {
  	QFileInfo inf( lst[i] );
	if ( !inf.isReadable() ) fail += lst[i];
  }

  if ( fail.size() ) {
  	QString msgtext = "Following files are not readable:\n";
	for ( i = 0; i < fail.size(); i++ )
		msgtext += fail[i] + "\n";
	msgtext += "Dictionary cannot be loaded";
  	QMessageBox::warning( this, "Pingvo Warning", msgtext,
		"Ok" );
	rebuildDictMenu();
	updateButtons();
	return;
  }
 
  if ( !status_bar ) {
  	status_bar = new QStatusBar( this );
	if ( layout() ) layout()->add( status_bar );
	status_bar->show();
	status_bar->message( QString("loading ") +
			dictionaries[dict_idx]->dname );
  }
  dloader.loadDictXML( dict_idx );
}

void IdxView::unloadDictXML( int dict_idx ) {
  if ( !status_bar ) {
  	status_bar = new QStatusBar( this );
	if ( layout() ) layout()->add( status_bar );
	status_bar->show();
	status_bar->message( QString("unloading ") +
			dictionaries[dict_idx]->dname );
  }
  dloader.unloadDictXML( dict_idx );
}

void IdxView::lookupByPrefix() {
  QListBoxItem *it;
  QString prefix;
  int dict;
  lookupEntry *entry;

  if ( dont_do_lookup ) return;

  prefix = le->text();
  entry = lookupNode( prefix, dict, true, true );
  if ( !entry ) return;
  
  while ( entry->word_number == -1 ) {
  	if ( !entry->dchilds_count ) return;
	entry = (lookupEntry*)(dictionaries[dict]->tree_start + entry->child_offs[0]);
  }
  
  it = dictionaries[dict]->itMap[entry->word_number];

  lb->setSelected( it, true );
  lb->setCurrentItem( it );
  lb->ensureCurrentVisible();
}

void IdxView::toggleDict1( bool on ) {
  if ( ign_dict_toggle || dict1 == -1 ) return;
  if ( on ) loadDictXML( dict1  ); else unloadDictXML( dict1 );
  dictMenu->setItemChecked( dictionaries[dict1]->menu_idx, on );
}

void IdxView::toggleDict2( bool on ) {
  if ( ign_dict_toggle || dict2 == -1 ) return;
  if ( on ) loadDictXML( dict2 ); else unloadDictXML( dict2 );
  dictMenu->setItemChecked( dictionaries[dict2]->menu_idx, on );
}

void IdxView::toggleDict3( bool on ) {
  if ( ign_dict_toggle || dict3 == -1 ) return;
  if ( on ) loadDictXML( dict3 ); else unloadDictXML( dict3 );
  dictMenu->setItemChecked( dictionaries[dict3]->menu_idx, on );
}

void IdxView::configure() {
  confDialog d(this);
  QComboBox *boxlist[] = { d.button1CB, d.button2CB, d.button3CB, 0 };
  QString labels[] = { QString::null, QString::null, QString::null };
  int *but_dicts[] = { &dict1, &dict2, &dict3, 0 };
  int rc, slot;
  unsigned int i, j;

  if ( dict1 != -1 && dictionaries.size() > (unsigned int)dict1 && dictionaries[dict1] )
  	labels[0] = dictionaries[dict1]->dname;
  if ( dict2 != -1 && dictionaries.size() > (unsigned int)dict2 && dictionaries[dict2] )
  	labels[1] = dictionaries[dict2]->dname;
  if ( dict3 != -1 && dictionaries.size() > (unsigned int)dict3 && dictionaries[dict3] )
  	labels[2] = dictionaries[dict3]->dname;

  for ( i = 0; i < dictionaries.size(); i++ ) {
  	if ( !dictionaries[i] ) continue;
	d.dictList->insertItem( dictionaries[i]->dname );
	for ( j = 0; boxlist[j]; j++ ) {
		boxlist[j]->insertItem( dictionaries[i]->dname );
		if ( dictionaries[i]->dname == labels[j] )
			boxlist[j]->setCurrentItem( boxlist[j]->count() - 1 );
	}
  }

  d.label1->setText( dict1B->text() );
  d.label2->setText( dict2B->text() );
  d.label3->setText( dict3B->text() );

  if ( confX != -1 && confY != -1 && confWidth != -1 && confHeight != -1 ) {
	d.move( confX, confY  );
	d.resize( confWidth, confHeight );
	d.show();
	qApp->processEvents();
	d.move( confX, confY  );
	d.resize( confWidth, confHeight );
  }

  rc = d.exec();

  confX = d.confX;
  confY = d.confY;
  confWidth = d.confWidth;
  confHeight = d.confHeight;
  
  if ( rc != QDialog::Accepted ) return;

  dict1B->setText( d.label1->text() );
  dict2B->setText( d.label2->text() );
  dict3B->setText( d.label3->text() );

  for ( i = 0; i < dictionaries.size(); i++ ) {
  	if ( !dictionaries[i] ) continue;
	for ( j = 0; j < d.dictList->count(); j++ )
	  	if ( dictionaries[i]->dname == d.dictList->text(j) )
			goto leave_dict;
	if ( dictionaries[i]->state != Dictionary::offline &&
	     dictionaries[i]->state != Dictionary::unloading )
		unloadDictXML( i );

	while ( dictionaries[i]->state != Dictionary::offline ) {
		qApp->processEvents();
		usleep( 100000 );
	}

	delete dictionaries[i];
	dictionaries[i] = 0;
leave_dict:;
  }

  QString ext;
  for ( i = 0; i < d.dictList->count(); i++ ) {
  	slot = -1;
  	for ( j = 0; j < dictionaries.size(); j++ ) {
		if ( dictionaries[j] && dictionaries[j]->dname == d.dictList->text(i) )
			goto already_have;
		if ( !dictionaries[j] && slot == -1 ) slot = j;
	}
	if ( slot == -1 ) {
		slot = dictionaries.size();
		dictionaries.resize( slot + 1 );
	}
	
	ext = d.dictList->text(i).right( 4 ).lower();
	if ( ext == ".xml" )
		dictionaries[slot] = new DictionaryXML;
	else if ( ext == ".dat" )
		dictionaries[slot] = new DictionaryDsl;
	else continue;

	initDictXML( dictionaries[slot], d.dictList->text(i) );
already_have:;
  }
  
  for ( i = 0; boxlist[i]; i++ )
  	for ( j = 0; j < dictionaries.size(); j++ )
		if ( dictionaries[j] && dictionaries[j]->dname ==
				boxlist[i]->currentText() ) {
			*but_dicts[i] = j;
			break;
		}

  for ( i = 0; but_dicts[i]; i++ )
	if ( *but_dicts[i] != -1 && (*but_dicts[i] >= dictionaries.size() ||
	     !dictionaries[*but_dicts[i]]) )
		*but_dicts[i] = -1;

  rebuildDictMenu();
}

void IdxView::rebuildDictMenu() {
  int it_idx = 0;

  ign_dict_toggle = true;
  dictMenu->clear();
  for ( unsigned int i = 0; i < dictionaries.size(); i++ ) {
  	if ( !dictionaries[i] ) continue;
  	it_idx = dictMenu->insertItem( dictionaries[i]->dname );
	dictionaries[i]->menu_idx = it_idx;
	if ( dictionaries[i]->state != Dictionary::offline )
		dictMenu->setItemChecked( it_idx, true );
  }
  ign_dict_toggle = false;
}

void IdxView::updateButtons() {
  bool on;

  ign_dict_toggle = true;
  if ( dict1 != -1 && dictionaries.size() > (unsigned int)dict1 && dictionaries[dict1] ) {
  	on = dictionaries[dict1]->state == Dictionary::ready ||
	     dictionaries[dict1]->state == Dictionary::loading;
  	dict1B->setOn( on );
  }
  if ( dict2 != -1 && dictionaries.size() > (unsigned int)dict2 && dictionaries[dict2] ) {
  	on = dictionaries[dict2]->state == Dictionary::ready ||
	     dictionaries[dict2]->state == Dictionary::loading;
  	dict2B->setOn( on );
  }
  if ( dict3 != -1 && dictionaries.size() > (unsigned int)dict3 && dictionaries[dict3] ) {
  	on = dictionaries[dict3]->state == Dictionary::ready ||
	     dictionaries[dict3]->state == Dictionary::loading;
  	dict3B->setOn( on );
  }
  ign_dict_toggle = false;
}

Article *IdxView::getArticle( int itDict, QString word ) {
  unsigned int i;
  Article *art;

  for ( i = 0; i < dictionaries.size(); i++ ) {
  	art = dictionaries[i]->getArticle( word );
  	if ( art ) {
		if ( itDict-- ) continue;
		return art;
	}
  }

  printf( "pingvo.cc: getArticle \"%s\" no such article\n", word.latin1() );
  return 0;
}

void IdxView::showCurrentArticle() {
  QListBoxItem *it = lb->selectedItem();
  if ( it ) showArticle( it );
}

ArticleView *IdxView::findUnpinnedBrowser() {
  for ( unsigned int i = 0; i < openedBrowsers.size(); i++ )
  	if ( !openedBrowsers[i]->pinB->isOn() ) return openedBrowsers[i];
  return 0;
}

QStringList IdxView::getDictList( QString word ) {
  lookupEntry *ent;
  QStringList lst;

  for ( unsigned int i = 0; i < dictionaries.size(); i++ ) {
  	ent = dictionaries[i]->lookupNode( word, false );
	if ( ent && ent->word_number != -1 )
		lst += dictionaries[i]->name;
  }

  return lst;
}

void IdxView::showArticle( QListBoxItem *lbItem ) {
  ArticleView *av;
  bool openedInNewWindow = false;

  Article *art;
  QString articleName = lbItem->text();
	
  articleName = lbItem->text();
  for ( unsigned int i = 0; i < openedBrowsers.size(); i++ )
	if ( openedBrowsers[i]->articleName == articleName ) {
		openedBrowsers[i]->raise();
		return;
	}

  art = getArticle( 0, articleName );
  if ( !art ) {
  	puts( "pingvo.cc: showArticle cannot get article" );
  	return;
  }
  av = findUnpinnedBrowser();
  if ( av ) {
	av->dictnum = 0;
	av->loadArticle( art, getDictList( articleName ) );
	delete art;
	av->raise();
	return;
  }
	
  av = new ArticleView();
  openedBrowsers.push_back( av );
  openedInNewWindow = true;

  connect( av, SIGNAL(geometryOnClose(ArticleView*,int,int,int,int)),
	SLOT(articleClosed(ArticleView*,int,int,int,int)) );
  connect( av->txtBrowser, SIGNAL(anchorClicked(const QString&, const QString&)),
	SLOT(processAnchorClick(const QString&, const QString&)) );
  connect( av, SIGNAL(changeDictRequest(ArticleView*,int)),
	SLOT(processChangeDict(ArticleView*,int)) );

  if ( geometryStack.size() ) {
	geometryItem it = geometryStack.back();
	geometryStack.resize( geometryStack.size() - 1 );

	av->move( it.x, it.y );
	av->resize( it.width, it.height );

	av->dictnum = 0;
	av->loadArticle( art, getDictList( articleName ) );
	qApp->processEvents();
		
	av->move( it.x, it.y );
	av->resize( it.width, it.height );
  }
  else {
	if ( openedBrowsers.size() == 1 ) av->move( defArtX, defArtY );
	av->resize( defArtWidth, defArtHeight );

	av->dictnum = 0;
	av->loadArticle( art, getDictList( articleName ) );
	qApp->processEvents();

	if ( openedBrowsers.size() == 1 ) av->move( defArtX, defArtY );
	av->resize( defArtWidth, defArtHeight );
  }

  delete art;
}

void IdxView::keyPressEvent( QKeyEvent *e ) {
  static bool skip_event = false;

  if ( skip_event ) return;

  if ( e->key() == Key_Up || e->key() == Key_Down ||
       e->key() == Key_PageUp || e->key() == Key_PageDown ) {
	skip_event = true;
  	qApp->sendEvent( lb, e );
	qApp->processEvents();
	skip_event = false;
	if ( !lb->currentText().isNull() ) {
		dont_do_lookup = true;
		le->setText( lb->currentText() );
		dont_do_lookup = false;
	}
	le->selectAll();
  }
}

void IdxView::moveEvent( QMoveEvent* ) {
  if ( !configured ) return;
  configChanged = true;
  if ( !configSaveTimer ) {
  	configSaveTimer = new QTimer( this );
	connect( configSaveTimer, SIGNAL(timeout()), SLOT(confSaveTimeout()) );
	configSaveTimer->start( 500, TRUE );
  }
}

void IdxView::closeEvent( QCloseEvent *e ) {
  configChanged = true;
  saveConfig();
  e->accept();
}

void IdxView::loadConfig() {
  QDomDocument doc;
  QDomElement docEl, e;
  QDomNode n, n2;
  QFile f( "config.xml" );
  bool idxgeom_initialized = false, artgeom_initialized = false;
  bool confgeom_initialized = false;
  bool geomstack_initialized = false, dictionaries_initialized = false;
  bool butlables_initialized = false;

  if ( !f.open( IO_ReadOnly ) ) return;
  doc.setContent( &f );
  n = doc.documentElement().firstChild();
  while ( !n.isNull() ) {
  	e = n.toElement();
	if ( e.tagName() == "IdxWinGeom" && !idxgeom_initialized ) {
		winX = e.attribute("x", "-1").toInt();
		winY = e.attribute("y", "-1").toInt();
		winW = e.attribute("w", "-1").toInt();
		winH = e.attribute("h", "-1").toInt();

		if ( winX != -1 && winY != -1 && winW != -1 && winH != -1 ) {
			move( winX, winY );
			resize( winW, winH );
		}

		idxgeom_initialized = true;
	}
	else if ( e.tagName() == "ConfWinGeom" && !confgeom_initialized ) {
		confX = e.attribute("x", "-1").toInt();
		confY = e.attribute("y", "-1").toInt();
		confWidth = e.attribute("w", "-1").toInt();
		confHeight = e.attribute("h", "-1").toInt();
		confgeom_initialized = true;
	}
	else if ( e.tagName() == "ArtWinGeom" && !artgeom_initialized ) {
		defArtX = e.attribute("x", "-1").toInt();
		defArtY = e.attribute("y", "-1").toInt();
		defArtWidth = e.attribute("w", "-1").toInt();
		defArtHeight = e.attribute("h", "-1").toInt();
		artgeom_initialized = true;
	}
	else if ( e.tagName() == "GeomStack" && !geomstack_initialized ) {
		int itm_num = e.attribute( "itm_num" ).toInt();
		geometryStack.resize( itm_num );
		n2 = e.firstChild();
		while ( !n2.isNull() ) {
			e = n2.toElement();
			if ( e.tagName() == "GeomItem" ) {
				int idx;
				geometryItem it;
				
				idx = e.attribute( "idx", "-1" ).toInt();
				if ( idx < 0 || idx >= itm_num ) {
					n2 = n2.nextSibling();
					continue;
				}
				it.x = e.attribute( "x" ).toInt();
				it.y = e.attribute( "y" ).toInt();
				it.width = e.attribute( "w" ).toInt();
				it.height = e.attribute( "h" ).toInt();

				geometryStack[idx] = it;
			}
			n2 = n2.nextSibling();
		}

		geomstack_initialized = true;
	}
	else if ( e.tagName() == "DictionariesList" && !dictionaries_initialized ) {
		int dict_idx = 0;

		n2 = e.firstChild();
		while ( !n2.isNull() ) {
			e = n2.toElement();
			if ( e.tagName() == "Dictionary" ) {
				if ( dict_idx >= (int)dictionaries.size() )
					dictionaries.resize( dict_idx + 1 );
				
				QString dictType = e.attribute( "type" );
				if ( dictType.isEmpty() || dictType == "xml" )
					dictionaries[dict_idx] = new DictionaryXML;
				else if ( dictType == "dsl" )
					dictionaries[dict_idx] = new DictionaryDsl;
					
				initDictXML( dictionaries[dict_idx],
					e.attribute( "xmlname" ) );

				if ( e.attribute( "load" ).toInt() == 1 )
					loadDictXML( dict_idx );
				dict_idx++;
			}
			n2 = n2.nextSibling();
		}
		dictionaries_initialized = true;
	}
	else if ( e.tagName() == "Buttons" && !butlables_initialized ) {
		int butnum;
		QPushButton *butlist[] = { dict1B, dict2B, dict3B };
		int *but_dict[] = { &dict1, &dict2, &dict3 };

		n2 = e.firstChild();
		while ( !n2.isNull() ) {
			e = n2.toElement();
			if ( e.tagName() != "Button" ) goto get_next_buttag;
			butnum = e.attribute( "id" ).toInt();
			if ( butnum < sizeof(butlist) / sizeof(QPushButton*) )
				butlist[butnum]->setText(e.attribute( "label" ));
			*but_dict[butnum] = e.attribute( "dict_num", "-1" ).toInt();
get_next_buttag:
			n2 = n2.nextSibling();
		}
		butlables_initialized = true;
	}
	else if ( e.tagName() == "SndArc" ) {
		sndarcpath = e.attribute( "path" );
	}
  	n = n.nextSibling();
  }
  f.close();

  updateButtons();
}

void IdxView::saveConfig() {
  QPoint p = pos();
  if ( !p.x() && !p.y() ) p = mapToGlobal(p);

  QFile f( "config.xml" );
  QDomDocument doc( "pingvo" );
  QDomElement docEl, e, e2;

  docEl = doc.createElement( "pingvo" );
  doc.appendChild( docEl );

  e = doc.createElement( "SndArc" );
  e.setAttribute( "path", sndarcpath );
  docEl.appendChild( e );

  e = doc.createElement( "IdxWinGeom" );
  e.setAttribute( "x", p.x() );
  e.setAttribute( "y", p.y() );
  e.setAttribute( "w", width() );
  e.setAttribute( "h", height() );
  docEl.appendChild( e );

  e = doc.createElement( "ConfWinGeom" );
  e.setAttribute( "x", confX );
  e.setAttribute( "y", confY );
  e.setAttribute( "w", confWidth );
  e.setAttribute( "h", confHeight );
  docEl.appendChild( e );

  e = doc.createElement( "ArtWinGeom" );
  e.setAttribute( "x", defArtX );
  e.setAttribute( "y", defArtY );
  e.setAttribute( "w", defArtWidth );
  e.setAttribute( "h", defArtHeight );
  docEl.appendChild( e );

  if ( geometryStack.size() || openedBrowsers.size() ) {
  	unsigned int i;
	int bsz;

	e = doc.createElement( "GeomStack" );
	bsz = openedBrowsers.size();
	e.setAttribute( "itm_num", geometryStack.size() + bsz );
	for ( i = 0; i < geometryStack.size(); i++ ) {
		e2 = doc.createElement( "GeomItem" );
		e2.setAttribute( "idx", i );
		e2.setAttribute( "x", geometryStack[i].x );
		e2.setAttribute( "y", geometryStack[i].y );
		e2.setAttribute( "w", geometryStack[i].width );
		e2.setAttribute( "h", geometryStack[i].height );
		e.appendChild( e2 );
	}
	for ( i = 0; i < openedBrowsers.size(); i++ ) {
		e2 = doc.createElement( "GeomItem" );
		e2.setAttribute( "idx", geometryStack.size() + i );
		e2.setAttribute( "x", openedBrowsers[bsz-i-1]->x() );
		e2.setAttribute( "y", openedBrowsers[bsz-i-1]->y() );
		e2.setAttribute( "w", openedBrowsers[bsz-i-1]->width() );
		e2.setAttribute( "h", openedBrowsers[bsz-i-1]->height() );
		e.appendChild( e2 );
	}
	docEl.appendChild( e );
  }

  e = doc.createElement( "DictionariesList" );
  for ( unsigned int i = 0; i < dictionaries.size(); i++ ) {
  	if ( !dictionaries[i] ) continue;
  	e2 = doc.createElement( "Dictionary" );
	e2.setAttribute( "xmlname", dictionaries[i]->dname );
	e2.setAttribute( "type", dictionaries[i]->type() );
	e2.setAttribute( "load", dictionaries[i]->state != Dictionary::offline? 1: 0 );
	e.appendChild( e2 );
  }
  docEl.appendChild( e );

  QPushButton *butlist[] = { dict1B, dict2B, dict3B };
  int but_dict[] = { dict1, dict2, dict3 };
  e = doc.createElement( "Buttons" );
  for ( int i = 0; i < sizeof(butlist) / sizeof(QPushButton*); i++ ) {
	e2 = doc.createElement( "Button" );
	e2.setAttribute( "id", i );
	e2.setAttribute( "dict_num", but_dict[i] );
	e2.setAttribute( "label", butlist[i]->text() );
	e.appendChild( e2 );
  }
  docEl.appendChild( e );
  
  if ( !f.open( IO_WriteOnly | IO_Truncate ) ) {
  	puts( "pingvo.cc config.xml open for writing failed" );
	return;
  }
  QTextStream stream( &f );
  doc.save( stream, 0 );
  f.close();

  configChanged = false;
}

void IdxView::confSaveTimeout() {
  delete configSaveTimer;
  configSaveTimer = 0;
  saveConfig();
}

void IdxView::articleClosed( ArticleView *av, int x, int y, int w, int h ) {
  if ( openedBrowsers.size() == 1 ) {
	defArtX = x;
	defArtY = y;
	defArtWidth = w;
	defArtHeight = h;
	saveConfig();
  }
  geometryItem it;
  it.x = x;
  it.y = y;
  it.width = w;
  it.height = h;
  geometryStack.push_back( it );

  for ( unsigned int i = 0; i < openedBrowsers.size(); i++ )
  	if ( openedBrowsers[i] == av ) {
		openedBrowsers.erase( openedBrowsers.begin() + i );
		return;
	}

  printf( "browsers list is inconsistent av=0x%08x, lst_size=%d\n",
  	av, openedBrowsers.size() );
}

void IdxView::xRecordEventsAvail() {
#ifdef __linux__
  recorder_process_replies();
#endif
}

void IdxView::processAnchorClick( const QString& word, const QString& ) {
  QListBoxItem *it;

  if ( word[0] == '#' ) {
	  QString sndfile = word.right(word.length() - 1);
	  if ( sndArc ) sndArc->SayWord( sndfile.ucs2() );
	  return;
  }

  it = lb->firstItem();
  while ( it ) {
  	if ( it->text() == word ) {
		showArticle( it );
		break;
	}
	it = it->next();
  }
}

void IdxView::processChangeDict( ArticleView *av, int dictnum ) {
  Article *art = getArticle( dictnum, av->articleName );
  if ( !art ) return;

  av->loadArticle( art, getDictList( av->articleName ) );
  delete art;
}

void IdxView::dictMenuAction( int id ) {
  if ( ign_dict_toggle ) return;
  for ( unsigned int i = 0; i < dictionaries.size(); i++ )
	if ( dictionaries[i]->menu_idx == id  ) {
		if ( !dictMenu->isItemChecked( id ) ) {
			dictMenu->setItemChecked( id, true );
			if ( dictionaries[i]->state != Dictionary::offline ) return;
  			loadDictXML( i );
			updateButtons();
		}
		else {
			dictMenu->setItemChecked( id, false );
			if ( dictionaries[i]->state == Dictionary::offline ||
			     dictionaries[i]->state == Dictionary::unloading )
				return;
			unloadDictXML( i );
			updateButtons();
		}
		break;
	}
}

void IdxView::customEvent( QCustomEvent *e ) {
  char c;
  QString ctext;
  QListBoxItem *it;
  lookupEntry *entry;

  if ( e->type() != QEvent::User + 2234 ) return;	// loaderEvent
  c = *(char*)e->data();

  if ( c == 1 ) {	// sent by dictionary loading thread when it finished
  	ctext = lb->currentText();
  	list <DictionaryLoader::loadedItem*>::iterator it = dloader.loaded.begin();
  	while ( it != dloader.loaded.end() ) {
		if ( (*it)->it )
			lb->insertItem( (*it)->it, (*it)->after );
		delete *it;
		++it;
	}
	dloader.loaded.clear();
  }
  else if ( c == 2 ) {	// send by dictionary unloading thread when it finished
  	ctext = lb->currentText();
  	list <IdxViewItem*>::iterator it = dloader.removed.begin();
	while ( it != dloader.removed.end() ) {
		delete *it;
		++it;
	}
	dloader.removed.clear();
  }
  else return;

  if ( !ctext.isEmpty() ) {
	int dict;

	entry = lookupNode( ctext, dict, true );
	if ( entry ) {
  		while ( entry->word_number == -1 ) {
		  	if ( !entry->dchilds_count ) return;
			entry = (lookupEntry*)(dictionaries[dict]->tree_start + entry->child_offs[0]);
		}
		it = dictionaries[dict]->itMap[entry->word_number];
		lb->setSelected( it, true );
		lb->setCurrentItem( it );
		lb->ensureCurrentVisible();
	}
  }

  if ( dloader.taskList.size() ) {
  	LoaderTask *task = dloader.taskList.front();

  	switch ( task->task ) {
	case LoaderTask::loadDict:
		status_bar->message( QString("loading ") +
			dictionaries[task->dict_idx]->dname );
		break;
	case LoaderTask::unloadDict:
		status_bar->message( QString("unloading ") +
			dictionaries[task->dict_idx]->dname );
		break;
	}
  }
  else if ( status_bar ) {
  	delete status_bar;
	status_bar = 0;
  }

  dloader.ackLock.lock();
  dloader.ack.wakeAll();
  dloader.ackLock.unlock();
}

PingvoTextBrowser::PingvoTextBrowser( QWidget *parent, const char *name ):
	QTextBrowser( parent, name ) {
}

void PingvoTextBrowser::setSource( const QString& /* s */ ) {
}

ArticleView::ArticleView() {
  setWFlags( getWFlags() | WDestructiveClose );
}

void ArticleView::loadArticle( Article *art, QStringList ldict ) {
  dictList->clear();
  dictList->insertStringList( ldict );
  dictList->setCurrentItem( dictnum );
 
  articleName = art->name;
 
  txtBrowser->mimeSourceFactory()->setFilePath(".");
  txtBrowser->setText( art->getRichText() );
  txtBrowser->setContentsPos( 0, 0 );
  show();
}

void ArticleView::closeEvent( QCloseEvent *e ) {
  QPoint p = pos();
  if ( !p.x() && !p.y() ) {
  	puts( "ArticleView doing mapToGlobal" );
  	p = mapToGlobal(p);
  }

  geometryOnClose( this, p.x(), p.y(), width(), height() );
  e->accept();
}

void ArticleView::dictNumCBChanged( int dict ) {
  dictnum = dict;
  changeDictRequest( this, dict );
}

void ArticleView::nextDictRequested() {
  if ( dictnum + 1 >= dictList->count() ) return;
  dictnum++;
  changeDictRequest( this, dictnum );
}

void ArticleView::prevDictRequested() {
  if ( !dictnum ) return;
  dictnum--;
  changeDictRequest( this, dictnum );
}

void confDialog::addDictionary() {
  QListBoxItem *it;
  QString fn = QFileDialog::getOpenFileName( QString::null, "*.xml;*.dat" );
  if ( fn.isEmpty() ) return;
  it = dictList->firstItem();
  while ( it ) {
  	if ( it->text() == fn ) return;
	it = it->next();
  }
  dictList->insertItem( fn );
  if ( !dictList->selectedItem() )
  	dictList->setSelected( dictList->firstItem(), true );
  button1CB->insertItem( fn );
  button2CB->insertItem( fn );
  button3CB->insertItem( fn );
}

void confDialog::removeDictionary() {
  QListBoxItem *it, *nxt;
  QComboBox *boxlist[] = { button1CB, button2CB, button3CB, 0 };
  QString text;
  int i, j;

  it = dictList->selectedItem();
  if ( !it ) return;
  text = it->text();
  if ( it->next() ) nxt = it->next();
  else if ( it->prev() ) nxt = it->prev();
  else nxt = 0;

  delete it;

  if ( nxt ) dictList->setSelected( nxt, true );
  for ( i = 0; boxlist[i]; i++ )
  	for ( j = 0; j < boxlist[i]->count(); j++ )
		if ( boxlist[i]->text(j) == text ) {
			boxlist[i]->removeItem(j);
			break;
		}
}

void confDialog::moveEntryUp() {
  int idx;
  QListBoxItem *it;

  it = dictList->selectedItem();
  if ( !it ) return;
  idx = dictList->index( it );
  if ( !idx ) return;
  dictList->setSelected( it, false );
  dictList->takeItem( it );
  idx--;
  dictList->insertItem( it, idx );
  dictList->setSelected( it, true );
}

void confDialog::moveEntryDown() {
  int idx;
  QListBoxItem *it;

  it = dictList->selectedItem();
  if ( !it ) return;
  idx = dictList->index( it );
  dictList->setSelected( it, false );
  dictList->takeItem( it );
  idx++;
  dictList->insertItem( it, idx );
  dictList->setSelected( it, true );
}

void confDialog::closeEvent( QCloseEvent *e ) {
  confX = x();
  confY = y();
  confWidth = width();
  confHeight = height();

  e->accept();
}

void confDialog::accept() {
  confX = x();
  confY = y();
  confWidth = width();
  confHeight = height();
  
  QDialog::accept();
}

void confDialog::reject() {
  confX = x();
  confY = y();
  confWidth = width();
  confHeight = height();

  QDialog::reject();
}

void magicReqHandler() {
  QClipboard *cb = qApp->clipboard();
  QString item = cb->text( QClipboard::Selection ).section( ' ', 0, 0 );

  idxview->le->setText( item.lower() );
  idxview->lookupByPrefix();
  if ( idxview->lb->currentText().lower() == item.lower() )
	idxview->showCurrentArticle();
}

int build_index() {
  IndexBuilderDsl indexer;
  puts( "building index" );
  indexer.build( "/mnt/hda2/ftp/pingvo/ScienceE.dat" );
  return 0;
}

int main( int argc, char **argv ) {
  QApplication qa( argc, argv );

#ifdef WIN32
#if 0
  AllocConsole();
  freopen( "CONOUT$", "w", stdout );
#endif
  AddFontResource( L"LvOemE16.fon" );
  setlocale( LC_ALL, "" );
#endif

  IdxView w;
// For ISoundArchive we need SndArc::path from config
// IdxView loads it at construction time
#ifdef __linux__
  QString dllname = "SoundArchive.dll.so";
  sndArc = (ISoundArchive*)NcCreateInstance( dllname.latin1(), &snddll );
#else
  QString dllname = "SoundArchive.dll";
  sndArc = (ISoundArchive*)NcCreateInstance( dllname.ucs2(), &snddll );
#endif
  if ( !sndArc )
	puts( "failed to create instance for ISoundArchive" );
  else if ( !sndarcpath.isNull() ) {
#ifdef __linux__
		int rc = sndArc->Open( sndarcpath.latin1() );
#else
		int rc = sndArc->Open( sndarcpath.ucs2() );
#endif
		if ( rc ) {
			puts( "failed to open archive" );
			sndArc = 0;
		}
  }

  idxview = &w;
  qa.setMainWidget( &w );

  qa.exec();

  if ( sndArc ) {
	  sndArc->Release();
	  NcDestroyInstance( &snddll );
  }

  return 0;
}
