#pragma once
#include <stdio.h>

#ifdef __linux__
#include <fcntl.h>
#include <unistd.h>

typedef char TCHAR;
#define _ASSERTE(a)
#define _tfopen(file, mode) fopen(file, mode)
#endif

class CSimpleFile
{
    FILE *m_hFile;

public:
    CSimpleFile() : m_hFile( NULL)
    {
    }

    ~CSimpleFile()
    {
        Close();
    }

    bool Open( const TCHAR * fileName, const TCHAR * mode)
    {
        _ASSERTE( m_hFile == NULL);
        m_hFile = _tfopen( fileName, mode);
        return (m_hFile != NULL);
    }

    void Close()
    {
        if (m_hFile != NULL)
        {
            fclose( m_hFile);
            m_hFile = NULL;
        }
    }

    int SeekCur( long offs)
    {
        return ::fseek( m_hFile, offs, SEEK_CUR);
    }

    int SeekEnd( long offs)
    {
        return ::fseek( m_hFile, offs, SEEK_END);
    }

    int SeekBeg( long offs)
    {
        return ::fseek( m_hFile, offs, SEEK_SET);
    }

    size_t Read( void *buffer, size_t bufSize)
    {
        return fread( buffer, 1, bufSize, m_hFile);
    }

    char ReadChar()
    {
        char ch = 0;
        fread( &ch, sizeof(ch), 1, m_hFile);
        return ch;
    }

    wchar_t ReadWChar()
    {
        uint16_t wch = 0;
        fread( &wch, sizeof(wch), 1, m_hFile);
        return wch;
    }

    unsigned long ReadDWord()
    {
        unsigned long dw = 0;
        fread( &dw, sizeof(dw), 1, m_hFile);
        return dw;
    }

    size_t Write( const void *buffer, size_t bufSize)
    {
        return fwrite( buffer, 1, bufSize, m_hFile);
    }

    long Tell()
    {
        return ftell( m_hFile);
    }

    bool operator!()
    {
        return (m_hFile == NULL);
    }

    operator FILE*()
    {
        return m_hFile;
    }
};
